# -*- coding: utf-8 -*-
"""Main tagger class
"""

import math
import os
import pickle

import numpy as np
import torch
from sklearn.base import BaseEstimator
from torch import nn, optim
from tqdm import tqdm

from .tagger_model import TaggerModel
from .utils import (DEVICE, batch_flow, build_vocabulary,
                    extract_char_features, extrat_entities, pad_seq,
                    prepare_sequence)


class Tagger(BaseEstimator):
    """Scikit-learn compatible Tagger
    """

    def __init__(self,
                 embedding_dim=100,
                 hidden_dim=200,
                 weight_decay=0.0,
                 epochs=100,
                 verbose=1,
                 batch_size=32,
                 device='auto',
                 embedding_dropout_p=0.0,
                 rnn_dropout_p=0.0,
                 bidirectional=True,
                 rnn_type='lstm',
                 num_layers=1,
                 optimizer='Adam',
                 momentum=0,
                 learning_rate=None,
                 learning_rate_decay=0,
                 embedding_trainable=True,
                 use_crf=True,
                 use_char='rnn',
                 char_max_len=50,
                 char_embedding_dim=30,
                 char_hidden_dim=50,
                 char_dropout_p=0.5,
                 char_bidirectional=True,
                 clip_grad_norm=None,
                 average_loss=False,
                 _model=None,
                 _optimizer=None,
                 _word_to_ix=None,
                 _ix_to_word=None,
                 _tag_to_ix=None,
                 _ix_to_tag=None,
                 _char_to_ix=None,
                 _ix_to_char=None):
        """init"""
        assert optimizer.lower() in ('sgd', 'adam')
        assert device.lower() in ('cpu', 'gpu', 'cuda', 'auto')
        self.params = {
            'embedding_dim': embedding_dim,
            'hidden_dim': hidden_dim,
            'weight_decay': weight_decay,
            'epochs': epochs,
            'verbose': verbose,
            'batch_size': batch_size,
            'device': device,
            'embedding_dropout_p': embedding_dropout_p,
            'rnn_dropout_p': rnn_dropout_p,
            'bidirectional': bidirectional,
            'rnn_type': rnn_type,
            'num_layers': num_layers,
            'optimizer': optimizer,
            'momentum': momentum,
            'learning_rate': learning_rate,
            'learning_rate_decay': learning_rate_decay,
            'embedding_trainable': embedding_trainable,
            'use_crf': use_crf,
            'use_char': use_char,
            'char_max_len': char_max_len,
            'char_embedding_dim': char_embedding_dim,
            'char_hidden_dim': char_hidden_dim,
            'char_dropout_p': char_dropout_p,
            'char_bidirectional': char_bidirectional,
            'clip_grad_norm': clip_grad_norm,
            'average_loss': average_loss,
        }
        self._model = _model
        self._optimizer = _optimizer
        self._word_to_ix = _word_to_ix
        self._ix_to_word = _ix_to_word
        self._tag_to_ix = _tag_to_ix
        self._ix_to_tag = _ix_to_tag
        self._char_to_ix = _char_to_ix
        self._ix_to_char = _ix_to_char

    def _get_learning_rate(self):
        """default learning rate"""
        optimizer = self.params['optimizer']
        learning_rate = self.params['learning_rate']
        if learning_rate is not None:
            return learning_rate
        if optimizer.lower() == 'sgd':
            return 1e-2
        if optimizer.lower() == 'adam':
            return 1e-3
        return 1e-3

    def get_params(self, deep=True):
        """Get params for scikit-learn compatible"""
        params = self.params
        if deep:
            params['_model'] = self._model.state_dict(
            ) if self._model is not None else None
            params['_optimizer'] = self._optimizer.state_dict() \
                if self._optimizer is not None else None
            params['_word_to_ix'] = self._word_to_ix
            params['_ix_to_word'] = self._ix_to_word
            params['_tag_to_ix'] = self._tag_to_ix
            params['_ix_to_tag'] = self._ix_to_tag
            params['_char_to_ix'] = self._char_to_ix
            params['_ix_to_char'] = self._ix_to_char
        return params

    def set_params(self, **parameters):
        """Set params for scikit-learn compatible"""
        for key, value in parameters.items():
            if key in self.params:
                self.params[key] = value
        return self

    def __getstate__(self):
        """Get state for pickle"""
        state = {
            'params':
            self.params,
            '_model':
            self._model.state_dict() if self._model is not None else None,
            '_optimizer':
            self._optimizer.state_dict()
            if self._optimizer is not None else None,
            '_word_to_ix':
            self._word_to_ix,
            '_ix_to_word':
            self._ix_to_word,
            '_tag_to_ix':
            self._tag_to_ix,
            '_ix_to_tag':
            self._ix_to_tag,
            '_char_to_ix':
            self._char_to_ix,
            '_ix_to_char':
            self._ix_to_char
        }
        return state

    def __setstate__(self, state):
        """Get state for pickle"""
        self.params = state['params']
        if state['_model'] is not None:
            self._word_to_ix = state['_word_to_ix']
            self._ix_to_word = state['_ix_to_word']
            self._tag_to_ix = state['_tag_to_ix']
            self._ix_to_tag = state['_ix_to_tag']
            self._char_to_ix = state['_char_to_ix']
            self._ix_to_char = state['_ix_to_char']
            self.apply_params()
            self._model.load_state_dict(state['_model'])
            self._optimizer.load_state_dict(state['_optimizer'])

    def _get_device(self):
        """Get device to predict or train"""
        device = self.params['device']
        if device == 'auto':
            return DEVICE
        if device in ('gpu', 'cuda'):
            return torch.device('cuda')
        return torch.device('cpu')

    def apply_params(self):
        """Apply params and build RNN-CRF model"""

        embedding_dim = self.params['embedding_dim']
        hidden_dim = self.params['hidden_dim']
        weight_decay = self.params['weight_decay']
        embedding_dropout_p = self.params['embedding_dropout_p']
        rnn_dropout_p = self.params['rnn_dropout_p']
        bidirectional = self.params['bidirectional']
        rnn_type = self.params['rnn_type']
        num_layers = self.params['num_layers']
        optimizer = self.params['optimizer']
        momentum = self.params['momentum']
        embedding_trainable = self.params['embedding_trainable']
        use_crf = self.params['use_crf']
        use_char = self.params['use_char']
        char_max_len = self.params['char_max_len']
        char_embedding_dim = self.params['char_embedding_dim']
        char_hidden_dim = self.params['char_hidden_dim']
        char_dropout_p = self.params['char_dropout_p']
        char_bidirectional = self.params['char_bidirectional']
        average_loss = self.params['average_loss']

        word_to_ix = self._word_to_ix
        tag_to_ix = self._tag_to_ix
        char_to_ix = self._char_to_ix

        model = TaggerModel(
            len(word_to_ix),
            tag_to_ix,
            embedding_dim,
            hidden_dim,
            num_layers=num_layers,
            bidirectional=bidirectional,
            device=self._get_device(),
            embedding_dropout_p=embedding_dropout_p,
            rnn_dropout_p=rnn_dropout_p,
            rnn_type=rnn_type,
            embedding_trainable=embedding_trainable,
            use_crf=use_crf,
            use_char=use_char,
            char_max_len=char_max_len,
            char_embedding_dim=char_embedding_dim,
            char_vocab_size=len(char_to_ix),
            char_hidden_dim=char_hidden_dim,
            char_dropout_p=char_dropout_p,
            char_bidirectional=char_bidirectional,
            average_loss=average_loss).to(self._get_device())

        if optimizer.lower() == 'adam':
            optimizer = optim.Adam(
                model.parameters(),
                lr=self._get_learning_rate(),
                weight_decay=weight_decay)
        elif optimizer.lower() == 'sgd':
            optimizer = optim.SGD(
                model.parameters(),
                lr=self._get_learning_rate(),
                weight_decay=weight_decay,
                momentum=momentum)

        self._model = model
        self._optimizer = optimizer

    def fit(self,
            X,
            y,
            X_dev=None,
            y_dev=None,
            patient_dev=None,
            save_best=None,
            pretrained_embedding=None,
            predict_batch_size=32):
        """Fit the model"""

        assert len(X) >= self.params['batch_size'], 'X must size >= batch_size'
        assert len(y) >= self.params['batch_size'], 'y must size >= batch_size'
        assert len(X) == len(y), 'X must size equal to y'

        # Autommatic build vocabulary
        vocabulary = build_vocabulary(X, y)
        self._char_to_ix = vocabulary['char_to_ix']
        self._ix_to_char = vocabulary['ix_to_char']
        if self._word_to_ix is None and self._tag_to_ix is None:
            self._word_to_ix = vocabulary['word_to_ix']
            self._ix_to_word = vocabulary['ix_to_word']
            self._tag_to_ix = vocabulary['tag_to_ix']
            self._ix_to_tag = vocabulary['ix_to_tag']
        elif self._word_to_ix is None:
            self._word_to_ix = vocabulary['word_to_ix']
            self._ix_to_word = vocabulary['ix_to_word']
        elif self._tag_to_ix is None:
            self._tag_to_ix = vocabulary['tag_to_ix']
            self._ix_to_tag = vocabulary['ix_to_tag']

        epochs = self.params['epochs']
        verbose = self.params['verbose']
        batch_size = self.params['batch_size']
        learning_rate_decay = self.params['learning_rate_decay']
        char_max_len = self.params['char_max_len']
        clip_grad_norm = self.params['clip_grad_norm']
        average_loss = self.params['average_loss']

        predict_batch_size = max(predict_batch_size, batch_size)

        word_to_ix = self._word_to_ix
        tag_to_ix = self._tag_to_ix
        char_to_ix = self._char_to_ix

        self.apply_params()
        model, optimizer = self._model, self._optimizer
        if pretrained_embedding is not None:
            model.load_embedding(pretrained_embedding)

        dev_best = float('inf')
        dev_best_round = 0
        # Make sure prepare_sequence from earlier in the LSTM section is loaded
        for epoch in range(epochs):
            model.train()
            model.zero_grad()
            lnrt = self._get_learning_rate()
            if learning_rate_decay > 0:
                lnrt = lnrt / (1 + epoch * learning_rate_decay)
                for param_group in optimizer.param_groups:
                    param_group['lr'] = lnrt
            pbar = range(len(X) // batch_size + 1)
            flow = batch_flow(
                X,
                y,
                word_to_ix,
                tag_to_ix,
                batch_size=batch_size,
                char_max_len=char_max_len,
                char_to_ix=char_to_ix)
            losses = []
            if verbose > 0:
                pbar = tqdm(pbar, ncols=0)
                pbar.set_description('epoch: {}/{} loss: {:.4f}'.format(
                    epoch + 1, epochs, 0))
            for _ in pbar:
                x_b, y_b, l_b, l_c, ll_c = next(flow)
                x_b = x_b.to(self._get_device())
                y_b = y_b.to(self._get_device())
                l_b = l_b.to(self._get_device())
                if l_c is not None:
                    l_c = l_c.to(self._get_device())
                    ll_c = ll_c.to(self._get_device())
                loss = model.compute_loss(
                    x_b, y_b, l_b, chars=l_c, charlens=ll_c)
                losses.append(loss.item())
                loss.backward()
                if clip_grad_norm is not None:
                    nn.utils.clip_grad_norm_(model.parameters(),
                                             clip_grad_norm)
                optimizer.step()
                model.zero_grad()
                if verbose > 0:
                    pbar.set_description(
                        'epoch: {}/{} loss: {:.4f} lr: {:.4f}'.format(
                            epoch + 1, epochs, np.mean(losses), lnrt))
            train_score = self.score(X, y, batch_size=predict_batch_size)
            dev_score = None
            if X_dev is None:
                if verbose > 0:
                    print('train: {:.4f}'.format(train_score))
            else:
                model.eval()
                with torch.no_grad():
                    dev_score = self.score(
                        X_dev, y_dev, batch_size=predict_batch_size)
                    flow = batch_flow(
                        X_dev,
                        y_dev,
                        word_to_ix,
                        tag_to_ix,
                        batch_size=predict_batch_size,
                        char_max_len=char_max_len,
                        char_to_ix=char_to_ix)
                    dev_losses = []
                    for _ in range(math.ceil(len(X_dev) / predict_batch_size)):
                        x_b, y_b, l_b, l_c, ll_c = next(flow)
                        x_b = x_b.to(self._get_device())
                        y_b = y_b.to(self._get_device())
                        l_b = l_b.to(self._get_device())
                        if l_c is not None:
                            l_c = l_c.to(self._get_device())
                            ll_c = ll_c.to(self._get_device())
                        loss = model.compute_loss(
                            x_b, y_b, l_b, chars=l_c, charlens=ll_c)
                        dev_losses.append(loss.item())
                    dev_loss = np.mean(dev_losses)
                    if not average_loss:
                        dev_loss /= (predict_batch_size / batch_size)

                if verbose > 0:
                    print('dev loss: {:.4f}, train f1: {:.4f}, dev f1: {:.4f}'.
                          format(dev_loss, train_score, dev_score))
                if isinstance(save_best, str):
                    if dev_loss < dev_best:
                        print('save best {:.4f} > {:.4f}'.format(
                            dev_best, dev_loss))
                        dev_best = dev_loss
                        dev_best_round = 0
                        save_dir = os.path.realpath(os.path.dirname(save_best))
                        if not os.path.exists(save_dir):
                            os.makedirs(save_dir)
                        with open(save_best, 'wb') as fobj:
                            pickle.dump(self, fobj)
                    else:
                        dev_best_round += 1
                        print('no better {:.4f} <= {:.4f} {}/{}'.format(
                            dev_best, dev_loss, dev_best_round, patient_dev))
                        if isinstance(patient_dev,
                                      int) and dev_best_round >= patient_dev:
                            return
                print()

    def predict(self, X, batch_size=None, verbose=0, max_limit=300):
        """Predict tags"""
        model = self._model
        word_to_ix = self._word_to_ix
        ix_to_tag = self._ix_to_tag
        char_to_ix = self._char_to_ix
        model.eval()
        if batch_size is None:
            batch_size = self.params['batch_size']
        char_max_len = self.params['char_max_len']
        # Check predictions after training
        data = list(enumerate(X))
        data = sorted(data, key=lambda t: len(t[1]), reverse=True)
        inds = [i for i, _ in data]
        X = [x for _, x in data]
        ret = [None] * len(X)
        with torch.no_grad():
            batch_total = math.ceil(len(X) / batch_size)
            pbar = range(batch_total)
            if verbose > 0:
                pbar = tqdm(pbar, ncols=0)
            for i in pbar:
                ind_batch = inds[i * batch_size:(i + 1) * batch_size]
                x_batch = X[i * batch_size:(i + 1) * batch_size]
                x_batch = [sent[:max_limit] for sent in x_batch]
                lens = [len(x) for x in x_batch]
                max_len = np.max(lens)

                char_batch = None
                char_len_batch = None
                if char_to_ix is not None:
                    char_batch, char_len_batch = extract_char_features(
                        x_batch, char_max_len, char_to_ix, max_len)

                x_batch = [
                    prepare_sequence(sent, word_to_ix) for sent in x_batch
                ]
                x_batch = [pad_seq(x, max_len) for x in x_batch]

                sent = torch.from_numpy(np.asarray(x_batch)).to(
                    self._get_device())
                lens = torch.Tensor(lens).long().to(self._get_device())
                char_batch = torch.from_numpy(np.asarray(char_batch)).to(
                    self._get_device())
                char_len_batch = torch.from_numpy(
                    np.asarray(char_len_batch,
                               dtype=np.int32)).to(self._get_device())
                _, predicts = model(sent, lens, char_batch, char_len_batch)
                for ind, tag_len, tags in zip(ind_batch, lens, predicts):
                    tags = [ix_to_tag[i] for i in tags[:tag_len]]
                    ret[ind] = tags
        return ret

    def score(self, X, y, batch_size=None, verbose=0, detail=False):
        """Calculate NER F1
        Based CONLL 2003 standard
        """

        def _get_sets():
            preds = self.predict(X, verbose=verbose, batch_size=batch_size)
            pbar = enumerate(zip(preds, y))
            if verbose > 0:
                pbar = tqdm(pbar, ncols=0, total=len(y))

            apset = []
            arset = []
            for i, (pred, y_true) in pbar:
                pset = extrat_entities(pred)
                rset = extrat_entities(y_true)
                for item in pset:
                    apset.append(tuple([i] + list(item)))
                for item in rset:
                    arset.append(tuple([i] + list(item)))
            return apset, arset

        apset, arset = _get_sets()
        pset = set(apset)
        rset = set(arset)
        inter = pset.intersection(rset)
        precision = len(inter) / len(pset) if pset else 1
        recall = len(inter) / len(rset) if rset else 1
        f1score = 0
        if precision + recall > 0:
            f1score = 2 * ((precision * recall) / (precision + recall))
        if detail:
            return precision, recall, f1score
        return f1score



# -*- coding: utf-8 -*-
"""Test NER"""

import os
import pickle
import random
import tempfile
import uuid

import numpy as np
import torch
from sklearn.model_selection import train_test_split
from torch_tagger import Tagger
from torch_tagger.utils import text_reader

SEED = 1337
random.seed(SEED)
torch.manual_seed(SEED)
torch.cuda.manual_seed(SEED)
np.random.seed(SEED)


def train_ner():
    """Train NER model"""
    x_data, y_data = text_reader('/tmp/msr_train.txt')
    tag = Tagger(
        embedding_dim=50,
        hidden_dim=50,
        weight_decay=1e-8,
        epochs=500,
        verbose=1,
        batch_size=20,
        device='auto',
        embedding_dropout_p=0.2,
        rnn_dropout_p=0.2,
        bidirectional=True,
        rnn_type='lstm',
        num_layers=1,
        optimizer='SGD',
        momentum=0.9,
        learning_rate=0.02,  # 0.02,
        learning_rate_decay=0.0,  # 0.05,
        embedding_trainable=True,
        use_crf=False,
        use_char='cnn',
        char_max_len=10,
        char_embedding_dim=50,
        char_hidden_dim=50,
        char_dropout_p=0.2,
        char_bidirectional=True,
        clip_grad_norm=None,
        average_loss=True)
    print('tag model', tag)

    tag_best_path = os.path.join(tempfile.gettempdir(),
                                 str(uuid.uuid4()) + '.pkl')
    print('tag_best_path', tag_best_path)
    x_train, x_test, y_train, y_test = train_test_split(
        x_data, y_data, test_size=0.2, random_state=SEED)
    tag.fit(x_train, y_train, x_test, y_test, save_best=tag_best_path)
    # with open(tag_best_path, 'wb') as fobj:
    #     pickle.dump(tag, fobj)
    test_ner(tag_best_path)
    generate_result_ner(tag_best_path, tag_best_path + '.txt')


def test_ner(model_path):
    """Test Segmentation model"""
    x_data, y_data = text_reader('/tmp/msr_test.txt')
    print('load model', model_path)
    tag = pickle.load(open(model_path, 'rb'))
    pre, rec, f1s = tag.score(x_data, y_data, verbose=1, detail=True)
    print('test precision {:.4f} recall {:.4f} f1 {:.4f}'.format(
        pre, rec, f1s))
    pred = tag.predict(x_data[:3])

    for i, indy in enumerate(y_data[:3]):
        print(indy)
        print(pred[i])
        print('-' * 30)


def generate_result_ner(model_path,
                        output_path,
                        test_input='/tmp/msr_test.txt'):
    """Generate file could evel by conll eval script"""
    x_data, _ = text_reader(test_input)
    print('load model', model_path)
    tag = pickle.load(open(model_path, 'rb'))
    pred = tag.predict(x_data)
    with open(output_path, 'w') as fobj:
        lines = open(test_input).read().split('\n\n')
        lines = [x.strip() for x in lines]
        lines = [x for x in lines if len(x) > 0]
        for line, line_pred in zip(lines, pred):
            line = [x.split()[0] for x in line.split('\n')]
            line_ret = ''
            for l, p in zip(line, line_pred):
                line_ret += l
                if p in ('E', 'S'):
                    line_ret += ' '
            fobj.write(line_ret + '\n')


if __name__ == '__main__':
    train_ner()



# -*- coding: utf-8 -*-
"""
.. module:: tagger_model

Based https://pytorch.org/tutorials/beginner/nlp/advanced_tutorial.html

Modified by InfinityFuture

"""

from torch import nn

from .char_cnn import CharCNN
from .char_rnn import CharRNN
from .decoder_crf import DecoderCRF
from .decoder_softmax import DecoderSoftmax
from .encoder_rnn import EncoderRNN
from .utils import sequence_mask


class TaggerModel(nn.Module):
    """Sequence Tagger Model
    """

    def __init__(self, vocab_size, tag_to_ix, embedding_dim, hidden_dim,
                 num_layers, bidirectional, embedding_dropout_p, rnn_dropout_p,
                 rnn_type, device, embedding_trainable, use_crf, use_char,
                 char_vocab_size, char_max_len, char_embedding_dim,
                 char_hidden_dim, char_dropout_p, char_bidirectional,
                 average_loss):
        """
        Embedding Random Init:

        [He et al.2015] Kaiming He, Xiangyu Zhang, Shaoqing
        Ren, and Jian Sun. 2015. Delving deep into recti-
        fiers: Surpassing human-level performance on ima-
        genet classification. In Proceedings of the IEEE In-
        ternational Conference on Computer Vision, pages
        1026–1034.
        """
        super(TaggerModel, self).__init__()

        if rnn_type not in ('lstm', 'gru'):
            raise Exception('Invalid rnn_type')

        self._use_crf = use_crf
        self._average_loss = average_loss

        self._encoder = EncoderRNN(
            vocab_size, tag_to_ix, embedding_dim, hidden_dim, num_layers,
            bidirectional, embedding_dropout_p, rnn_dropout_p, rnn_type,
            device, embedding_trainable, use_char, char_hidden_dim)

        if use_crf:
            self._decoder = DecoderCRF(tag_to_ix, device)
        else:
            self._decoder = DecoderSoftmax()

        self._char_features = None
        if isinstance(use_char, str):
            if use_char.lower() == 'rnn':
                self._char_features = CharRNN(
                    char_vocab_size, char_max_len, char_embedding_dim,
                    char_hidden_dim, char_dropout_p, char_bidirectional,
                    rnn_type, device, embedding_trainable)
            elif use_char.lower() == 'cnn':
                self._char_features = CharCNN(char_vocab_size, char_max_len,
                                              char_embedding_dim,
                                              char_hidden_dim, char_dropout_p,
                                              device, embedding_trainable)

    def load_embedding(self, pretrained_embedding):
        """Load other pre-trained embedding vectors"""
        self._encoder.load_embedding(pretrained_embedding)

    def compute_loss(self, sentence, tags, lengths, chars=None, charlens=None):
        """Compute Train Loss"""
        char_feats = None
        if chars is not None and self._char_features is not None:
            char_feats = self._char_features(chars, charlens)
        encoder_output = self._encoder(
            sentence, lengths, char_feats=char_feats)
        loss = self._decoder.compute_loss(encoder_output, tags, lengths)
        if self._average_loss:
            batch_size = sentence.size(0)
            loss /= batch_size
        return loss

    def forward(self, sentence, lengths, chars=None, charlens=None):
        """Main forward function, predict only"""

        char_feats = None
        if chars is not None and self._char_features is not None:
            char_feats = self._char_features(chars, charlens)

        # Get the emission scores from the BiLSTM
        rnn_feats = self._encoder(sentence, lengths, char_feats=char_feats)
        seq_len = rnn_feats.size(1)
        masks = sequence_mask(lengths, seq_len)

        if self._use_crf:
            # CRF
            # Find the best path, given the features.
            return self._decoder(rnn_feats, lengths, masks)
        # Softmax
        return self._decoder(rnn_feats)



# -*- coding: utf-8 -*-
"""Test NER with GloVe

Download https://nlp.stanford.edu/projects/glove/
"""

import os
import random
import tempfile
import uuid

import numpy as np
import torch
from torch_tagger import Tagger
from torch_tagger.utils import PAD_TAG, UNK_TAG, build_vocabulary, text_reader

from .ner import generate_result_ner, test_ner

SEED = 222
random.seed(SEED)
torch.manual_seed(SEED)
torch.cuda.manual_seed(SEED)
np.random.seed(SEED)


def read_glove(train_txt='/tmp/train.txt',
               glove_path='tests/glove.6B/glove.6B.100d.txt',
               dim=100):
    """Reade GloVe vector"""

    x_data, y_data = text_reader(train_txt)
    x_val, y_val = text_reader('/tmp/valid.txt')
    x_test, y_test = text_reader('/tmp/test.txt')
    vocabulary = build_vocabulary(x_data + x_val + x_test,
                                  y_data + y_val + y_test)

    def random_vec():
        return np.random.uniform(
            -np.sqrt(3. / dim), np.sqrt(3. / dim), size=(1, dim))

    def zero_vec():
        return np.zeros((1, dim))

    predefine_vecs = {}
    with open(glove_path, 'r') as fobj:
        lines = fobj.read().split('\n')
        lines = [x.strip() for x in lines]
        lines = [x for x in lines if x.strip()]
        for line in lines:
            line = line.split()
            if len(line) > dim:
                word = line[0]
                vec = np.array([float(i) for i in line[1:]])
                vec = vec.reshape(1, dim)
                predefine_vecs[word] = vec
    vecs = [
        zero_vec(),
        random_vec(),
    ]
    word_to_ix = {PAD_TAG: 0, UNK_TAG: 1}
    perfect = 0
    fuzzy = 0
    none = 0
    for word in vocabulary['word_to_ix']:
        if word not in word_to_ix:
            indx = len(word_to_ix)
            word_to_ix[word] = indx
            if word in predefine_vecs:
                vecs.append(predefine_vecs[word])
                perfect += 1
            elif word.lower() in predefine_vecs:
                vecs.append(predefine_vecs[word.lower()])
                fuzzy += 1
            elif word.upper() in predefine_vecs:
                vecs.append(predefine_vecs[word.upper()])
                fuzzy += 1
            else:
                vecs.append(random_vec())
                none += 1

    ix_to_word = {v: k for k, v in word_to_ix.items()}
    embedding = np.concatenate(vecs)

    print('perfect: {}, fuzz: {}, none: {}'.format(perfect, fuzzy, none))

    return word_to_ix, ix_to_word, embedding


def train_ner():
    """Train NER model"""
    x_data, y_data = text_reader('/tmp/train.txt')
    x_val, y_val = text_reader('/tmp/valid.txt')
    word_to_ix, ix_to_word, embedding = read_glove()
    dropout = 0.5
    tag = Tagger(
        embedding_dim=100,
        hidden_dim=200,
        weight_decay=1e-8,
        epochs=100,
        verbose=1,
        batch_size=10,
        device='auto',
        embedding_dropout_p=dropout,
        rnn_dropout_p=dropout,
        bidirectional=True,
        rnn_type='lstm',
        num_layers=1,
        optimizer='SGD',
        learning_rate=0.015,
        learning_rate_decay=0.05,
        embedding_trainable=True,
        use_crf=True,
        use_char='lstm',
        char_max_len=50,
        char_embedding_dim=30,
        char_hidden_dim=50,
        char_dropout_p=dropout,
        char_bidirectional=True,
        clip_grad_norm=None,
        _word_to_ix=word_to_ix,
        _ix_to_word=ix_to_word)
    print('tag model', tag)

    tag_best_path = os.path.join(tempfile.gettempdir(),
                                 str(uuid.uuid4()) + '.pkl')
    print('tag_best_path', tag_best_path)

    tag.fit(
        x_data,
        y_data,
        x_val,
        y_val,
        patient_dev=20,
        save_best=tag_best_path,
        pretrained_embedding=embedding)

    score = tag.score(x_data, y_data, verbose=1)
    print('train score', score)
    test_ner(tag_best_path)
    generate_result_ner(tag_best_path, tag_best_path + '.txt')


if __name__ == '__main__':
    train_ner()



# -*- coding: utf-8 -*-
"""Test utils class"""

import unittest

from torch_tagger.utils import pad_seq


class TestUtil(unittest.TestCase):
    def test_pad_seq(self):
        r = pad_seq([1, 2, 3], 5)
        self.assertEqual(r, [1, 2, 3, 0, 0])


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""Utils tools for tagger
Created by InfinityFuture
"""

import re
from collections import Counter

import numpy as np
import torch
from sklearn.utils import shuffle

DEVICE = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
START_TAG = '<START>'
STOP_TAG = '<STOP>'
PAD_TAG = '<PAD>'
UNK_TAG = '<UNK>'


def prepare_sequence(seq, to_ix) -> torch.LongTensor:
    """Convert sequence to torch variable"""
    idxs = [to_ix[w] if w in to_ix else to_ix[UNK_TAG] for w in seq]
    return idxs


def normalize_word(word):
    """normalize number"""
    word = re.sub(r'\d', '0', word)
    return word


def text_reader(path):
    """Read a text file, and return data
    data should follow this format:

    I O
    want O
    to O
    New B-City
    York I-City
    """
    with open(path, 'r') as fobj:
        parts = fobj.read().split('\n\n')
        parts = [part.strip() for part in parts]
        parts = [part for part in parts if len(part) > 0]
    assert parts, 'text file empty "{}"'.format(path)

    x_data = []
    y_data = []
    for part in parts:
        lines = part.split('\n')
        lines = [line.split() for line in lines]
        words = [x[0] for x in lines]
        tags = [x[-1] for x in lines]
        words = [normalize_word(word) for word in words]
        x_data.append(words)
        y_data.append(tags)
        assert len(words) == len(tags), \
            'line "{}" and "{}" do not match "{}" vs "{}"'.format(
                len(words), len(tags), words, tags)
    return x_data, y_data


def build_vocabulary(x_data: list, y_data: list, limit=1) -> dict:
    """ Use data to build vocabulary"""
    sentence_word = Counter()
    sentence_word_char = Counter()
    tags_word = Counter()
    for sentence in x_data:
        sentence_word.update(sentence)
    for tags in y_data:
        tags_word.update(tags)

    word_to_ix = {PAD_TAG: 0, UNK_TAG: 1}
    for word, count in sentence_word.items():
        sentence_word_char.update(list(word))
        if word not in word_to_ix and count >= limit:
            indx = len(word_to_ix)
            word_to_ix[word] = indx

    char_to_ix = {PAD_TAG: 0, UNK_TAG: 1}

    for char, count in sentence_word_char.items():
        if count >= limit:
            indx = len(char_to_ix)
            char_to_ix[char] = indx

    ix_to_char = {v: k for k, v in char_to_ix.items()}

    ix_to_word = {v: k for k, v in word_to_ix.items()}

    tag_to_ix = {}

    tag_words = [PAD_TAG] + list(tags_word.keys()) + [START_TAG, STOP_TAG]
    for tag in tag_words:
        indx = len(tag_to_ix)
        tag_to_ix[tag] = indx

    ix_to_tag = {v: k for k, v in tag_to_ix.items()}

    return {
        'word_to_ix': word_to_ix,
        'ix_to_word': ix_to_word,
        'tag_to_ix': tag_to_ix,
        'ix_to_tag': ix_to_tag,
        'char_to_ix': char_to_ix,
        'ix_to_char': ix_to_char
    }


def pad_seq(seq: list, max_len: int, pad=0) -> list:
    """Padding data to max_len length"""
    if len(seq) < max_len:
        return seq + [pad] * (max_len - len(seq))
    return seq


def extract_char_features(x_batch, char_max_len, char_to_ix, max_len):
    """extract char features"""
    char_batch = []
    char_len_batch = []
    for xseq in x_batch:
        sentence_char = []
        len_char = []
        for char in xseq:
            if re.match(r'^[\u4e00-\u9fff]+$', char):
                # Process Chinese
                from pypinyin import lazy_pinyin
                char = ' '.join([x[0] for x in lazy_pinyin(char)])
            char = char[:char_max_len]
            len_char.append(len(char))
            char = prepare_sequence(char, char_to_ix)
            char = pad_seq(char, char_max_len)
            sentence_char.append(char)
        while len(sentence_char) < max_len:
            sentence_char.append([0] * char_max_len)
        while len(len_char) < max_len:
            len_char.append(0)
        char_batch.append(sentence_char)
        char_len_batch.append(len_char)
    char_batch = np.asarray(char_batch)
    char_len_batch = np.asarray(char_len_batch)
    return char_batch, char_len_batch


def batch_flow(x_data: list,
               y_data: list,
               word_to_ix: dict,
               tag_to_ix: dict,
               char_to_ix: dict = None,
               char_max_len: int = None,
               batch_size: int = 32,
               sample_shuffle=True,
               max_limit=300):
    """Automatic generate batch data"""
    assert len(x_data) >= batch_size, 'len(x_data) > batch_size'
    assert len(x_data) == len(y_data), \
        'len(x_data) == len(y_data), {} != {}'.format(len(x_data), len(y_data))

    x_batch, y_batch = [], []
    inds = list(range(len(x_data)))
    if sample_shuffle:
        inds = shuffle(inds)
    while True:
        for ind in inds:
            if len(x_batch) == batch_size:
                len_batch = [len(t) for t in x_batch]
                max_len = np.max(len_batch)

                char_batch = None
                char_len_batch = None
                if char_to_ix is not None:
                    char_batch, char_len_batch = extract_char_features(
                        x_batch, char_max_len, char_to_ix, max_len)

                x_batch = [prepare_sequence(x, word_to_ix) for x in x_batch]
                y_batch = [prepare_sequence(y, tag_to_ix) for y in y_batch]
                x_batch = [pad_seq(x, max_len) for x in x_batch]
                y_batch = [pad_seq(y, max_len) for y in y_batch]

                batches = list(
                    zip(x_batch, y_batch, len_batch, char_batch,
                        char_len_batch))
                batches = sorted(batches, key=lambda x: x[2], reverse=True)
                (x_batch, y_batch, len_batch, char_batch,
                 char_len_batch) = zip(*batches)

                tcx, tcy, tcl, tcc, tccl = (torch.from_numpy(
                    np.asarray(x_batch)), torch.from_numpy(
                        np.asarray(y_batch)),
                                            torch.from_numpy(
                                                np.asarray(len_batch)),
                                            torch.from_numpy(
                                                np.asarray(char_batch)),
                                            torch.from_numpy(
                                                np.asarray(
                                                    char_len_batch,
                                                    dtype=np.int32)))
                x_batch, y_batch = [], []
                yield tcx, tcy, tcl, tcc, tccl

            x_batch.append(x_data[ind][:max_limit])
            y_batch.append(y_data[ind][:max_limit])


def sequence_mask(lens: torch.Tensor,
                  max_len: int = None) -> torch.FloatTensor:
    """InfinityFutures: This function is copy from:

    https://github.com/epwalsh/pytorch-crf

    The author is epwalsh, and its license is MIT too

    Compute sequence mask.

    Parameters

    lens : torch.Tensor
        Tensor of sequence lengths ``[batch_size]``.
    max_len : int, optional (default: None)
        The maximum length (optional).

    Returns

    torch.ByteTensor
        Returns a tensor of 1's and 0's of size ``[batch_size x max_len]``.
    """
    batch_size = lens.size(0)

    if max_len is None:
        max_len = lens.max().item()

    ranges = torch.arange(0, max_len, device=lens.device).long()
    ranges = ranges.unsqueeze(0).expand(batch_size, max_len)
    ranges = torch.autograd.Variable(ranges)

    lens_exp = lens.unsqueeze(1).expand_as(ranges)
    mask = ranges < lens_exp

    return mask.float()


def extrat_entities(seq: list) -> list:
    """Extract entities from a sequences

    ---
    input: ['B', 'I', 'I', 'O', 'B', 'I']
    output: [(0, 3, ''), (4, 6, '')]
    ---
    input: ['B-loc', 'I-loc', 'I-loc', 'O', 'B-per', 'I-per']
    output: [(0, 3, '-loc'), (4, 6, '-per')]
    """
    ret = []
    start_ind, start_type = -1, None
    for i, tag in enumerate(seq):
        if tag.startswith('S'):
            ret.append((i, i + 1, tag[1:]))
            start_ind, start_type = -1, None
        if tag.startswith('B') or tag.startswith('O'):
            if start_ind >= 0:
                ret.append((start_ind, i, start_type))
                start_ind, start_type = -1, None
        if tag.startswith('B'):
            start_ind = i
            start_type = tag[1:]
    if start_ind >= 0:
        ret.append((start_ind, len(seq), start_type))
        start_ind, start_type = -1, None
    return ret



# -*- coding: utf-8 -*-
"""Test NER"""

import os
import pickle
import random
import tempfile
import uuid

import numpy as np
import torch
from torch_tagger import Tagger
from torch_tagger.utils import text_reader

SEED = 1337
random.seed(SEED)
torch.manual_seed(SEED)
torch.cuda.manual_seed(SEED)
np.random.seed(SEED)


def train_ner():
    """Train NER model"""
    x_data, y_data = text_reader('/tmp/train.txt')
    x_val, y_val = text_reader('/tmp/valid.txt')
    tag = Tagger(
        embedding_dim=100,
        hidden_dim=100,
        weight_decay=1e-8,
        epochs=100,
        verbose=1,
        batch_size=10,
        device='auto',
        embedding_dropout_p=0.5,
        rnn_dropout_p=0.5,
        bidirectional=True,
        rnn_type='lstm',
        num_layers=1,
        optimizer='SGD',
        learning_rate=0.015,
        learning_rate_decay=0.05,
        embedding_trainable=True,
        use_crf=False,
        use_char='cnn',
        char_max_len=50,
        char_embedding_dim=30,
        char_hidden_dim=50,
        char_dropout_p=0.5,
        char_bidirectional=True,
        clip_grad_norm=None,
    )
    print('tag model', tag)

    tag_best_path = os.path.join(tempfile.gettempdir(),
                                 str(uuid.uuid4()) + '.pkl')
    print('tag_best_path', tag_best_path)
    tag.fit(x_data, y_data, x_val, y_val, save_best=tag_best_path)
    test_ner(tag_best_path)
    generate_result_ner(tag_best_path, tag_best_path + '.txt')


def test_ner(model_path):
    """Test NER model"""
    x_data, y_data = text_reader('/tmp/test.txt')
    print('load model', model_path)
    tag = pickle.load(open(model_path, 'rb'))
    pre, rec, f1s = tag.score(x_data, y_data, verbose=1, detail=True)
    print('test precision {:.4f} recall {:.4f} f1 {:.4f}'.format(
        pre, rec, f1s))
    pred = tag.predict(x_data[:3])

    for i, indy in enumerate(y_data[:3]):
        print(indy)
        print(pred[i])
        print('-' * 30)


def generate_result_ner(model_path, output_path='/tmp/test_pred.txt'):
    """Generate file could evel by conll eval script"""
    x_data, y_data = text_reader('/tmp/test.txt')
    print('load model', model_path)
    tag = pickle.load(open(model_path, 'rb'))
    pred = tag.predict(x_data)
    with open(output_path, 'w') as fobj:
        for xbch, ybch, pbch in zip(x_data, y_data, pred):
            for ind, x_one in enumerate(xbch):
                fobj.write('{} {} {}\n'.format(x_one, ybch[ind], pbch[ind]))


if __name__ == '__main__':
    train_ner()



# -*- coding: utf-8 -*-
"""Test Entry"""

import os
import unittest

from sklearn.model_selection import GridSearchCV
from torch_tagger import Tagger
from torch_tagger.utils import text_reader

CURRENT = os.path.realpath(os.path.dirname(__file__))
PATH = os.path.join(CURRENT, 'train.txt')


class TestGrid(unittest.TestCase):
    def test_grid(self):
        """Test GridSearch"""
        x_data, y_data = text_reader(PATH)
        parameters = {
            'bidirectional': (True, False),
            'rnn_type': ('lstm', 'gru'),
            'num_layers': (1, 2),
            'use_char': ('cnn', 'rnn', None),
            'use_crf': (True, False),
        }
        x_train = x_data + x_data + x_data
        y_train = y_data + y_data + y_data
        tag = Tagger(batch_size=2, epochs=1, verbose=0, device='cpu')
        clf = GridSearchCV(tag, parameters, verbose=1, n_jobs=1)
        clf.fit(x_train, y_train)
        self.assertIsInstance(clf.cv_results_['mean_test_score'][0], float)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""Test Entry"""

import unittest

if __name__ == '__main__':
    testsuite = unittest.TestLoader().discover('.', pattern='test*.py')
    unittest.TextTestRunner(verbosity=1).run(testsuite)
